/*-
 * wrapper library for long paths exceeding PATH_MAX
 * 
 * Copyright (c) 2013 UPLEX Nils Goroll Systemoptimierung
 * All rights reserved
 *
 * Author: Nils Goroll <nils.goroll@uplex.de>
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * == IMPLEMENTATION NOTES ==
 *
 * UNIMPLEMENTED
 * - ftw, nftw
 * - copylist
 * - stdio.h stuff: fopen, freopen
 * - acl
 * - Check for O_SEARCH ?
 *
 * TODO
 * - use pathconf instead of fixed MAX_PATH
 * - cache intermediate fds
 * - cache getpath results
 *
 * Notes on lf64:
 * for the intermediate filedescriptors, we always use openat64 unless _LP64 we
 * can do so because we never hand out these fds - Xatl/Xatl2 always close them
 *
 * the lib gets compiled with _FILE_OFFSET_BITS==32 for 32 bit builds so we can
 * provide both variants. longpath.h will provide the appropriate redefinitions
 * depending on our consumer
 *
 * Notes on PATH_MAX:
 * Ref: http://pubs.opengroup.org/onlinepubs/009695399/basedefs/limits.h.html
 * Maximum number of bytes in a pathname, including the terminating null character.
 *
 */

#include "../config.h"

#define _ATFILE_SOURCE

#if defined(_LP64)
#define	_FILE_OFFSET_BITS	64
#define six4(x) x
#else
#define	_FILE_OFFSET_BITS	32
#define six4(x) x ## 64
#endif
//// no use to undef these - solaris  sys/feature_tests.h sets to 1
//#undef  _LARGEFILE_SOURCE
//#undef  _LARGEFILE64_SOURCE

#include <longpath/longpath.h>

#ifdef HAVE_ASSERT_H
#include <assert.h>
#else
#define assert(EX) ((void)0)
#endif

// XXX conditional includes incomplete

// string functions
#if defined(HAVE_STRING_H)
#include <string.h>
#endif

#ifdef HAVE_LIMITS_H
#include <limits.h>
#endif
#ifdef HAVE_ERRNO_H
#include <errno.h>
#endif
#ifdef HAVE_STDIO_H
#include <stdio.h>
#endif

// varargs for open
#if   defined(HAVE_SYS_VARARGS_H)
#include <sys/varargs.h>
#elif defined(HAVE_STDARG_H)
#include <stdarg.h>
#endif


// opendir
#ifdef HAVE_DIRENT_H
#include <dirent.h>
#endif

// linux xattr
#ifdef HAVE_FGETXATTR
#include <sys/xattr.h>
#endif

static int _openatl(int, const char *, int, mode_t mode);
#if !defined(_LP64)
static int _openatl64(int, const char *, int, mode_t mode);
#endif


/*
 * close an intermediate file descriptor, restoring errno if necessary
 */
static inline void
close_ifd(int ifd) {
	const int s = errno;

	assert(ifd >= 0);
	assert(ifd != (int)AT_FDCWD);

	if (close(ifd) == -1)
		errno = s;
}

#ifdef USE_CHDIR_FALLBACK

/*
 * define NEED_CHDIR when using these
 */
static inline void restore_cwd(int *fd);
static int chdir_im(const char *path, const char **remaining, int limit);

#define Xl(funcn, path, ...)						\
	int r;								\
	int ocwd = open(".", O_RDONLY|O_NDELAY|O_LARGEFILE);		\
	const char *remaining;						\
									\
	if (ocwd == -1)							\
		return -1;						\
									\
	if (chdir_im(path, &remaining, 0) == -1) {			\
		restore_cwd(&ocwd);					\
		return -1;						\
	}								\
									\
	path += (remaining - path);					\
									\
	r = funcn(path, ##__VA_ARGS__);					\
									\
	restore_cwd(&ocwd);						\
	return r

#endif /* USE_CHDIR_FALLBACK */

/*
 * oatl_im = open at long intermediate
 *
 * for a path longer than PATH_MAX, open intermediate directories and return an
 * fd to the last directory component required for colling *at functions on the
 * remaining path component
 *
 * filedes can be AT_FDCWD
 * openatl_interm will not close filedes
 *
 * return:
 *
 *        -1	for errors, errno will be set
 * filedes	if path is NULL or shorter than PATH_MAX
 * otherwise	new open filedescriptor
 *
 * The caller must close the returned fd unless it is equal to filedes
 */

static int
oatl_im(int filedes, const char *path, const char **remaining) {
	char		interm[PATH_MAX];
	const char	*p, *sl;
	size_t		l;
	int		fd;

	/*
	 * work with functions on NULL paths like fchownat
	 * - at any rate we can leave error handling with the system function
	 */
	if (path == NULL) {
		*remaining = NULL;
		return filedes;
	}

	l = strlen(path);

	if (l < PATH_MAX) {
		*remaining = path;
		return filedes;
	}

	p = path;
	fd = filedes;
	while (l >= PATH_MAX) {
		int nfd;

		assert(l == strlen(p));

		/* find last slash for strlen(component) < PATH_MAX */
		sl = p + PATH_MAX - 1;

		while (sl > p) {
			if (*sl == '/')
				goto found;
			sl--;
		}

		errno = ENAMETOOLONG;
		goto err;

	  found:
		assert((sl - p) < PATH_MAX);
		strncpy(interm, p, (sl - p));
		interm[sl - p] = '\0';

		nfd = six4(openat)(fd, interm,
		    O_RDONLY|O_NDELAY|O_LARGEFILE, 0);

		if (nfd == -1)
			goto err;

		if (fd != filedes)
			close_ifd(fd);

		fd = nfd;
		l -= (sl - p) + 1;
		p  = sl + 1;
	}

	assert(fd != filedes);
	*remaining = p;
	return fd;

    err:
	if (fd != filedes)
		close_ifd(fd);
	return -1;
}

// long path *at function
#define Xatl(funcn, filedes, path, ...)					\
	int		__ifd, __fd;					\
	const char	*__rest;					\
									\
	__ifd = oatl_im(filedes, path, &__rest);			\
									\
	if (__ifd == -1)						\
		return -1;						\
									\
	__fd = funcn(__ifd, __rest, ##__VA_ARGS__);			\
									\
	if (__ifd != filedes)						\
		close_ifd(__ifd);					\
									\
	return __fd

// long path *at function with two paths
#define Xatl2(funcn, fd1, p1, fd2, p2, ...)				\
	int		__ifd1, __ifd2, __fd;				\
	const char	*__rest1;					\
	const char	*__rest2;					\
									\
	__ifd1 = oatl_im(fd1, p1, &__rest1);				\
	if (__ifd1 == -1)						\
		return -1;						\
									\
	__ifd2 = oatl_im(fd2, p2, &__rest2);				\
	if (__ifd2 == -1) {						\
		__fd = -1;						\
		goto close_ifd1;					\
	}								\
									\
	__fd = funcn(__ifd1, __rest1, __ifd2, __rest2, ##__VA_ARGS__);	\
									\
	if (__ifd2 != fd2)						\
		close_ifd(__ifd2);					\
									\
  close_ifd1:								\
	if (__ifd1 != fd1)						\
		close_ifd(__ifd1);					\
									\
	return __fd

// long path f* function - make mode an argument?
#define Xfl(path, flags, err, r, funcn, ...)				\
	int		__ifd;						\
									\
	__ifd = six4(_openatl)(AT_FDCWD, (path),			\
	    O_RDONLY|O_NDELAY|O_LARGEFILE|(flags), 0);			\
									\
	if (__ifd == -1)						\
		return (err);						\
									\
	r = funcn(__ifd, ##__VA_ARGS__);				\
									\
	close_ifd(__ifd)

/* OPEN */
static int
_openatl(int filedes, const char *path, int oflag, mode_t mode) {
	Xatl(openat, filedes, path, oflag, mode);
}

#if !defined(_LP64)
static int
_openatl64(int filedes, const char *path, int oflag, mode_t mode) {
	Xatl(openat64, filedes, path, oflag, mode);
}
#endif

#define Openatl(bits)						\
	mode_t mode = 0;					\
	if (oflag & O_CREAT) {					\
		va_list arg;					\
		va_start(arg, oflag);				\
		mode = va_arg(arg, mode_t);			\
		va_end(arg);					\
	}							\
								\
	return _openatl ## bits (filedes, path, oflag, mode)

int
openatl(int filedes, const char *path, int oflag, ...) {
	Openatl(/*32*/);
}

#if !defined(_LP64)
int
openatl64(int filedes, const char *path, int oflag, ...) {
	Openatl(64);
}
#endif

#define Openl(bits)						\
	mode_t mode = 0;					\
	if (oflag & O_CREAT) {					\
		va_list arg;					\
		va_start(arg, oflag);				\
		mode = va_arg(arg, mode_t);			\
		va_end(arg);					\
	}							\
								\
	return _openatl ## bits (AT_FDCWD, path, oflag, mode)


int
openl(const char *path, int oflag, ...) {
	Openl(/*32*/);
}
#if !defined(_LP64)
int
openl64(const char *path, int oflag, ...) {
	Openl(64);
}
#endif

/* CREAT */

int
creatl(const char *path, mode_t mode) {
	return _openatl(AT_FDCWD, path, O_WRONLY|O_CREAT|O_TRUNC , mode);
}
#if !defined(_LP64)
int
creatl64(const char *path, mode_t mode) {
	return _openatl64(AT_FDCWD, path, O_WRONLY|O_CREAT|O_TRUNC , mode);
}
#endif

/* ATTROPEN */
/* Solaris specific ? */

#ifdef HAVE_ATTROPEN
#define Attropenl(bits)							\
        int fd, afd;							\
        va_list ap;							\
									\
	fd = openl ## bits (path, O_RDONLY|O_NONBLOCK);				\
	if (fd == -1)							\
		return -1;						\
									\
	va_start(ap, oflag);						\
	afd = _openatl ## bits (fd, apath, oflag|O_XATTR,		\
	    va_arg(ap, mode_t));					\
	va_end(ap);							\
									\
	close_ifd(fd);							\
	return afd

int
attropenl(const char *path, const char *apath, int oflag, ...) {
	Attropenl(/*32*/);
}
#if !defined(_LP64)
int
attropenl64(const char *path, const char *apath, int oflag, ...) {
	Attropenl(64);
}
#endif
#endif /* HAVE_ATTROPEN */

/* GETXATTR / LGETXATTR */
/* untested */
#ifdef HAVE_FGETXATTR

#ifndef HAVE_FSETXATTR
#error Assuming fsetxattr to exist when fgetxattr exists
#endif

ssize_t
getxattrl(const char *path, const char *name, void *value, size_t size) {
	ssize_t r;
	Xfl(path, 0, -1,
	    r, fgetxattr, name, value, size);
	return r;
}
ssize_t
lgetxattrl(const char *path, const char *name, void *value, size_t size) {
	ssize_t r;
	Xfl(path, AT_SYMLINK_NOFOLLOW, -1,
	    r, fgetxattr, name, value, size);
	return r;
}
int
setxattrl(const char *path, const char *name, const void *value, size_t size, int flags) {
	int r;
	Xfl(path, 0, -1,
	    r, fsetxattr, name, value, size, flags);
	return r;
}
int
lsetxattrl(const char *path, const char *name, const void *value, size_t size, int flags) {
	int r;
	Xfl(path, AT_SYMLINK_NOFOLLOW, -1,
	    r, fsetxattr, name, value, size, flags);
	return r;
}
#endif /* HAVE_FGETXATTR */


/* LINK */

#ifdef HAVE_LINKAT
int
linkatl(int fd1, const char *p1, int fd2, const char *p2, int flags) {
	Xatl2(linkat, fd1, p1, fd2, p2, flags);
}

int
linkl(const char *p1, const char *p2) {
	return linkatl(AT_FDCWD, p1, AT_FDCWD, p2, 0);
}
#else /* HAVE_LINKAT */
#ifndef USE_CHDIR_FALLBACK
#warning linkat missing - link* calls on long paths may fail
int
linkl(const char *p1, const char *p2) {
	return link(p1, p2);
}

#else /* USE_CHDIR_FALLBACK */
#warning linkat missing - link* calls on long paths may fail (trying prefix match)
static int
strprefixlen(const char *s1, const char *s2) {
	int i = 0;
	while((*s1 != '\0') && (*s1 == *s2)) {
		s1++; s2++; i++;
	}

	return i;
}

int
linkl(const char *p1, const char *p2) {
	int l = strprefixlen(p1, p2);

	if (l > 0) {
		int r;
		int ocwd = open(".", O_RDONLY|O_NDELAY|O_LARGEFILE);
		const char *remaining;

		if (ocwd == -1)
			return -1;

		if (chdir_im(p1, &remaining, l) == -1) {
			restore_cwd(&ocwd);
			return -1;
		}

		l = remaining - p1;

		p1 += l;
		p2 += l;

		r = link(p1, p2);

		restore_cwd(&ocwd);
		return r;
	}

	return link(p1, p2);
}
#endif /* USE_CHDIR_FALLBACK */
#endif /* HAVE_LINKAT */

/** UNLINK **/

int
unlinkatl(int fd, const char *p, int flags) {
	Xatl(unlinkat, fd, p, flags);
}

int
unlinkl(const char *p) {
	return unlinkatl(AT_FDCWD, p, 0);
}

/** SYMLINK **/
#ifdef HAVE_SYMLINKAT
/* slightly differnt signature - damn */
int
symlinkatl(const char *dest, int filedes, const char *path) {
	int		ifd, r;
	const char	*rest;

	ifd = oatl_im(filedes, path, &rest);

	if (ifd == -1)
		return -1;

	r = symlinkat(dest, ifd, rest);

	if (ifd != filedes)
		close_ifd(ifd);

	return r;
}
int
symlinkl(const char *dest, const char *path) {
	return symlinkatl(dest, AT_FDCWD, path);
}
#else /* HAVE_SYMLINKAT */
#ifndef USE_CHDIR_FALLBACK
#warning symlinkat missing - symlink* calls on long paths may fail
int
symlinkl(const char *dest, const char *path) {
	return symlink(dest, path);
}
#else /* USE_CHDIR_FALLBACK */
int
symlinkl(const char *dest, const char *path) {
	int r;
	int ocwd = open(".", O_RDONLY|O_NDELAY|O_LARGEFILE);
	const char *remaining;

	if (ocwd == -1)
		return -1;

	if (chdir_im(path, &remaining, 0) == -1) {
		restore_cwd(&ocwd);
		return -1;
	}

	path += (remaining - path);

	r = symlink(dest, path);

	restore_cwd(&ocwd);
	return r;
}

#endif /* USE_CHDIR_FALLBACK */
#endif /* HAVE_SYMLINKAT */


/** READLINK **/

#ifdef HAVE_READLINKAT
int
readlinkatl(int fd, const char *name, char *buf, size_t count) {
	Xatl(readlinkat, fd, name, buf, count);
}

int
readlinkl(const char *name, char *buf, size_t count) {
	return readlinkatl(AT_FDCWD, name, buf, count);
}
#else /* HAVE_READLINKAT */
#ifndef USE_CHDIR_FALLBACK
#warning readlinkat missing - link* calls on long paths may fail
int
readlinkl(const char *name, char *buf, size_t count) {
	return readlink(name, buf, count);
}
#else /* USE_CHDIR_FALLBACK */
#undef  NEED_CHDIR
#define NEED_CHDIR
int
readlinkl(const char *name, char *buf, size_t count) {
	Xl(readlink, name, buf, count);
}
#endif /* USE_CHDIR_FALLBACK */
#endif /* HAVE_READLINKAT */


/** CHOWN **/

int
fchownatl(int fd, const char *name, uid_t uid, gid_t gid, int flags) {
	Xatl(fchownat, fd, name, uid, gid, flags);
}

int
chownl(const char *name, uid_t uid, gid_t gid) {
	return fchownatl(AT_FDCWD, name, uid, gid, 0);
}
int
lchownl(const char *name, uid_t uid, gid_t gid) {
	return fchownatl(AT_FDCWD, name, uid, gid, AT_SYMLINK_NOFOLLOW);
}

/** CHMOD **/
int
chmodl(const char *path, mode_t mode) {
	int fd;
	Xfl(path, 0, -1,
	    fd, fchmod, mode);
	return fd;
}

/* MKNOD */

#ifdef HAVE_MKNODAT
int
mknodatl(int fd, const char *name, mode_t mode, dev_t dev) {
	Xatl(mknodat, fd, name, mode, dev);
}
int
mknodl(const char *name, mode_t mode, dev_t dev) {
	return mknodatl(AT_FDCWD, name, mode, dev);
}
#else /* HAVE_MKNODAT */
#ifndef USE_CHDIR_FALLBACK
#warning mknodat missing - link* calls on long paths may fail
int
mknodl(const char *name, mode_t mode, dev_t dev) {
	return mknod(name, mode, dev);
}
#else /* USE_CHDIR_FALLBACK */
#undef  NEED_CHDIR
#define NEED_CHDIR
int
mknodl(const char *name, mode_t mode, dev_t dev) {
	Xl(mknod, name, mode, dev);
}
#endif /* USE_CHDIR_FALLBACK */
#endif /* HAVE_MKNODAT */

/* MKFIFO */

#ifdef HAVE_MKFIFOAT
int
mkfifoatl(int fd, const char *name, mode_t mode) {
	Xatl(mkfifoat, fd, name, mode);
}
int
mkfifol(const char *name, mode_t mode) {
	return mkfifoatl(AT_FDCWD, name, mode);
}
#else /* HAVE_MKFIFOAT */
#ifndef USE_CHDIR_FALLBACK
#warning mkfifoat missing - link* calls on long paths may fail
int
mkfifol(const char *name, mode_t mode) {
	return mkfifo(name, mode);
}
#else /* USE_CHDIR_FALLBACK */
#undef  NEED_CHDIR
#define NEED_CHDIR
int
mkfifol(const char *name, mode_t mode) {
	Xl(mkfifo, name, mode);
}
#endif /* USE_CHDIR_FALLBACK */
#endif /* HAVE_MKFIFOAT */


/* STAT */
int
fstatatl(int fd, const char *name, struct stat *sb, int flags) {
	Xatl(fstatat, fd, name, sb, flags);
}

int
statl(const char *name, struct stat *sb) {
	return fstatatl(AT_FDCWD, name, sb, 0);
}

int
lstatl(const char *name, struct stat *sb) {
	return fstatatl(AT_FDCWD, name, sb, AT_SYMLINK_NOFOLLOW);
}

#if !defined(_LP64)
int
fstatatl64(int fd, const char *name, struct stat64 *sb, int flags) {
	Xatl(fstatat64, fd, name, sb, flags);
}

int
statl64(const char *name, struct stat64 *sb) {
	return fstatatl64(AT_FDCWD, name, sb, 0);
}

int
lstatl64(const char *name, struct stat64 *sb) {
	return fstatatl64(AT_FDCWD, name, sb, AT_SYMLINK_NOFOLLOW);
}
#endif

/* RMDIR */

int
rmdiratl(int fd, const char *name) {
	return unlinkatl(fd, name, AT_REMOVEDIR);
}

int
rmdirl(const char *name) {
	return rmdiratl(AT_FDCWD, name);
}

/* MKDIR */

#ifdef HAVE_MKDIRAT
int
mkdiratl(int fd, const char *name, mode_t mode) {
	Xatl(mkdirat, fd, name, mode);
}

int
mkdirl(const char *name,  mode_t mode) {
	return mkdiratl(AT_FDCWD, name, mode);
}
#else /* HAVE_MKDIRAT */
#ifndef USE_CHDIR_FALLBACK
#warning mkdirat missing - link* calls on long paths may fail
int
mkdirl(const char *name,  mode_t mode) {
	return mkdir(name, mode);
}
#else /* USE_CHDIR_FALLBACK */
#undef  NEED_CHDIR
#define NEED_CHDIR
int
mkdirl(const char *name,  mode_t mode) {
	Xl(mkdir, name, mode);
}
#endif /* USE_CHDIR_FALLBACK */
#endif /* HAVE_MKDIRAT */

/* RENAME */

int
renameatl(int fd1, const char *p1, int fd2, const char *p2) {
	Xatl2(renameat, fd1, p1, fd2, p2);
}

int
renamel(const char *p1, const char *p2) {
	return renameatl(AT_FDCWD, p1, AT_FDCWD, p2);
}


/*
 * MK*TEMP*
 *
 * there is no mk*tempat, so we need to implement it outselves
 */

/*
 * common code for all the mk*temp* - structure taken from glibc, which is
 * Copyright (C) 1991-2001, 2006, 2007, 2009 Free Software Foundation, Inc.
 */

enum temp_kind {
	T_NOCREATE,
	T_FILE,
	T_DIR
#if !defined(_LP64)
	,T_FILE64
#endif
};

/*
 * return val:
 * -1 or valid fd
 *
 * AT_FDCWD is returned to signal success for T_DIR and T_NOCREATE
 */

static inline int
_mkxtempl(char *template, int slen, int flags, enum temp_kind kind) {
	size_t			len;
	int			ifd, fd;
	char     		*rest;
	char			*save, *r, *spos;
	struct six4(stat)	tstat;

	if (slen < 0 || template == NULL || *template == '\0') {
		errno = EINVAL;
		return -1;
	}

	len = strlen(template);
	spos = NULL;
	save = NULL;

	/* check input */
	{
		char	*end, *tmp;

		end = &template[len - 1];
		/* no trailing slashes */
		if (*end == '/') {
			errno = EINVAL;
			return -1;
		}
		tmp = end;
		while (--tmp >= template) {
			if (*tmp == '/') {
				/*
				 * if we have a slash, the suffix must be within
				 * the last path element
				 */
				if (slen >= (end - tmp)) {
					errno = EINVAL;
					return -1;
				}
			}
		}
		tmp = end - slen;
		if (*tmp == 'X')
			spos = tmp + 1;
	}

	ifd = oatl_im(AT_FDCWD, template, (const char **)&rest);
	if (ifd == -1)
		return -1;

	// save = strdupa(rest);
	save = malloc(strlen(rest) + 1);
	if (! save) {
		errno = ENOMEM;
		goto err;
	}
	strcpy(save, rest);

	do {
		/*
		 * XXX: as we don't have mktempat, mktemp will check the wrong
		 * directory for long paths. this does not really matter because
		 * we check for exclusive createion / existance afterwards, but
		 * is less efficient.
		 */
		if (spos) {
			char	s = *spos;

			/* temp. trim of the suffix for mktemp */
			*spos = '\0';
			r = mktemp(rest);
			*spos = s;

			if (*r == '\0')
				goto err;
		}

		switch (kind) {
		case T_DIR:
		{
#ifdef HAVE_MKDIRAT
			if (mkdirat(ifd, rest, 0700) == 0) {
				fd = AT_FDCWD;
				goto out;
			}
#else /* HAVE_MKDIRAT */
#ifndef USE_CHDIR_FALLBACK
#warning mkdirat missing - mkdtemp* calls on long paths may fail
			if (mkdtemp(template)) {
				fd = AT_FDCWD;
				goto out;
			}
			goto err;
#else /* USE_CHDIR_FALLBACK */
			if (ifd == AT_FDCWD) {
				if (mkdir(rest, 0700) == 0) {
					fd = AT_FDCWD;
					goto out;
				}
			} else {
				int ocwd;
				ocwd = open(".", O_RDONLY|O_NDELAY|O_LARGEFILE);
				if (ocwd == -1)
					goto err;
				if (fchdir(ifd) == -1)
					goto err;
				if (mkdir(rest, 0700) == 0) {
					restore_cwd(&ocwd);
					fd = AT_FDCWD;
					goto out;
				}
				restore_cwd(&ocwd);
			}
#endif /* USE_CHDIR_FALLBACK */
#endif /* HAVE_MKDIRAT */

			break; /* retry */
		}
		case T_FILE:
			fd = openat(ifd, rest, O_CREAT|O_EXCL|O_RDWR|flags, 0600);

			if (fd != -1)
				goto out;

			break; /* retry */
#if !defined(_LP64)
		case T_FILE64:
			fd = openat64(ifd, rest, O_CREAT|O_EXCL|O_RDWR|flags, 0600);

			if (fd != -1)
				goto out;

			break; /* retry */
#endif
		case T_NOCREATE:
		{
			int serrno = errno;
			if (six4(fstatatl)(ifd, rest, &tstat, AT_SYMLINK_NOFOLLOW) == -1) {
				if (errno != ENOENT)
					goto err;

				assert(errno == ENOENT);
				errno = serrno;
				fd = AT_FDCWD;
				goto out;
			}
			break; /* retry */
		}
		default:
			assert(! "invalid kind in _mkxtempl");
		}

		if ((! spos) || (errno != EEXIST))
			goto err;
		(void) strcpy(rest, save);
	} while (1);

	assert (! "NOTREACHED");

  err:
	fd = -1;
  out:
	if (save)
		free(save);
	if (ifd != (int)AT_FDCWD)
		close_ifd(ifd);
	return fd;
}

/*
 * Quote solaris man page:
 *   The mkdtemp() function can set errno to the same  values  as
 *   mkdir(2).
 * so we don't need to save errno
 */

char *
mktempl(char *template) {
	int serrno = errno;
	if (_mkxtempl(template, 0, 0, T_NOCREATE) == -1)
		return NULL;
	errno = serrno;
	return template;
}

char *
mkdtempl(char *template) {
	if (_mkxtempl(template, 0, 0, T_DIR) == -1)
		return NULL;
	return template;
}

int
mkstempl(char *template) {
	return _mkxtempl(template, 0, 0, T_FILE);
}

int
mkstempsl(char *template, int slen) {
	return _mkxtempl(template, slen, 0, T_FILE);
}

int
mkostempl(char *template, int flags) {
	return _mkxtempl(template, 0, flags, T_FILE);
}

int
mkostempsl(char *template, int slen, int flags) {
	return _mkxtempl(template, slen, flags, T_FILE);
}

#if !defined(_LP64)
int
mkstempl64(char *template) {
	return _mkxtempl(template, 0, 0, T_FILE64);
}

int
mkstempsl64(char *template, int slen) {
	return _mkxtempl(template, slen, 0, T_FILE64);
}

int
mkostempl64(char *template, int flags) {
	return _mkxtempl(template, 0, flags, T_FILE64);
}

int
mkostempsl64(char *template, int slen, int flags) {
	return _mkxtempl(template, slen, flags, T_FILE64);
}
#endif

/* UTIMES */

/*
 * (f)utimesat is special and strange, so map them all to utimensat, which is
 * the new syscall anyway
 */

/*
 * http://pubs.opengroup.org/onlinepubs/9699919799/functions/futimens.html#tag_16_181
 *
 * tv_nsec UTIME_NOW : same as NULL tv
 * tv_nsec UTIME_OMIT: don't set
 */

int
utimensatl(int fd, const char *path, const struct timespec times[2], int flag) {
	Xatl(utimensat, fd, path, times, flag);
}


static inline void
tv2ts(struct timespec **tsp, const struct timeval tv[2]) {
	if (tv == NULL) {
		*tsp = NULL;
	} else {
		struct timespec *ts = *tsp;
		assert(ts);
		ts[0].tv_sec  = tv[0].tv_sec;
		ts[0].tv_nsec = tv[0].tv_usec * 1000;
		ts[1].tv_sec  = tv[1].tv_sec;
		ts[1].tv_nsec = tv[1].tv_usec * 1000;
	}
}

/*
 * futimesat is special
 * - if PATH is NULL, update is applied to the fd
 * - if PATH is relative, PATH is relative to fd
 *
 * but because utimensatl does not handle PATH == NULL, we need
 * to pass to futimens
 *
 */
int
futimesatl(int filedes, const char *path, const struct timeval tv[2]) {
	struct timespec ts[2];
	struct timespec *tsp = ts;

	tv2ts(&tsp, tv);

	if (path == NULL)
		return futimens(filedes, tsp);

	return utimensatl(filedes, path, tsp, 0);
}

int
utimesl(const char *path, const struct timeval tv[2]) {
	struct timespec ts[2];
	struct timespec *tsp = ts;

	tv2ts(&tsp, tv);

	return utimensatl(AT_FDCWD, path, tsp, 0);
}

int
lutimesl(const char *path, const struct timeval tv[2]) {
	struct timespec ts[2];
	struct timespec *tsp = ts;

	tv2ts(&tsp, tv);

	return utimensatl(AT_FDCWD, path, tsp, AT_SYMLINK_NOFOLLOW);
}

int
utimel(const char *path, const struct utimbuf *times) {
	if (times) {
		struct timespec ts[2];

		ts[0].tv_sec  = times->actime;
		ts[0].tv_nsec = 0;

		ts[1].tv_sec  = times->modtime;
		ts[1].tv_nsec = 0;

		return utimensatl(AT_FDCWD, path, ts, 0);
	}

	return utimensatl(AT_FDCWD, path, NULL, 0);
}

/* OPENDIR */
DIR *
opendirl(const char *dirname) {
	int fd;
	/* almost identical to Xfl, but we must not close the fd */
	fd = _openatl(AT_FDCWD, dirname, O_RDONLY|O_NDELAY|O_LARGEFILE, 0);

	if (fd == -1)
		return NULL;

	return fdopendir(fd);
}

#ifdef NEED_CHDIR

/*
 * this function will always try to chdir even if path is shorter than PATH_MAX
 * so it can be used on a common prefix of two paths
 */
static int
chdir_im(const char *path, const char **remaining, int limit) {
	char		interm[PATH_MAX];
	const char	*p, *sl;
	size_t		l;

	/*
	 * work with functions on NULL paths like fchownat
	 * - at any rate we can leave error handling with the system function
	 */
	if (path == NULL) {
		*remaining = NULL;
		return 0;
	}

	l = strlen(path);
	if ((limit > 0) && (l > limit))
		l = limit;

	p = path;
	do {
		/* find last slash within limit */
		sl = p + ((l < PATH_MAX) ? l : (PATH_MAX - 1));

		while (sl > p) {
			if (*sl == '/')
				goto found;
			sl--;
		}

		goto out;

	  found:
		assert((sl - p) < PATH_MAX);
		strncpy(interm, p, (sl - p));
		interm[sl - p] = '\0';

		if (chdir(interm) == -1)
			goto err;

		l -= (sl - p) + 1;
		p  = sl + 1;
	} while (1);

  out:
	*remaining = p;
	return 0;

  err:
	return -1;
}

static inline void
restore_cwd(int *fd) {
	assert(*fd >= 0);
	assert(*fd != (int)AT_FDCWD);

	if (fchdir(*fd) == -1) {
		/*
		 * XXX: This really must not happen, we have
		 *      changed the processes' state and can't
		 *      restore it. Any better ideas?
		 */
		assert(! "liblongpath restore_cwd");
		perror("liblongpath restore_cwd");
		exit(1);
	}

	close_ifd(*fd);
	*fd = -1;
}

#else /* NEED_CHDIR */

static int chdir_im(const char *path, const char **remaining, int limit) {
	assert(! "not used");
	(void)path;
	(void)remaining;
	(void)limit;
	return -1;
}
static inline void restore_cwd(int *fd) {
	assert(! "not used");
	(void)fd;
}
void *_dummy1 = chdir_im;
void *_dummy2 = restore_cwd;

#endif /* NEED_CHDIR */
